/*
Jana's Energy Saber
	by Lach

v2 (bottomslope is kill)
*/

rawset(_G, "JanaEnergySaber", {})

local JES = JanaEnergySaber

local function valid(mo)
	return mo and mo.valid
end

// To prevent duplicate freeslots
local function SafeFreeslot(...)
	local function CheckSlot(item) // this function deliberately errors when a freeslot does not exist
		if _G[item] == nil // this will error by itself for states and objects
			error() // this forces an error for sprites, which do actually return nil for some reason
		end
	end
	for _, item in ipairs({...})
		if pcall(CheckSlot, item)
			print("\131NOTICE:\128 " .. item .. " was not allocated, as it already exists.")
		else
			freeslot(item)
		end
	end
end

// ------
// sounds
// ------

if sfxinfo[sfx_s3ka5].caption == ""
	sfxinfo[sfx_s3ka5].caption = "/"
end

for i = 1, 3
	SafeFreeslot("sfx_jaatk"..i)
	sfxinfo[sfx_jaatk1 + i - 1].caption = "Saber"
end

// -------
// sprites
// -------

SafeFreeslot("SPR_JSWD", "SPR_SSWD", "SPR_SLCB", "SPR_SLC2", "SPR_SLC3")

// ------
// states
// ------

JES.AttackAction = function(mo, var1, var2)
	local player = mo.player
	
	if not player
		return
	end
	
	local frame = mo.frame & FF_FRAMEMASK
	local jes = player.janaEnergySaber
	
	// change tics depending on frame
	if frame > 0
		mo.tics = frame + var2
	end
	
	// spawn a projectile if using a combo attack
	if frame == C
	and (JES.InComboState(player) or jes and jes.flags & JES.FLAGS.CHARGEATTACK)
		JES.SpawnSaberBeam(player)
	end
	
	// set panim
	player.panim = PA_DASH // LOL I'm starting to get tired of PA_DASH hacks
end

JES.AttackEndAction = function(mo)
	local player = mo.player
	
	if not player
		return
	end
	
	local jana = player.jana
	local jes = player.janaEnergySaber
	
	if jana and jana.dashing and jes.lastAttackState == S_JANA_DASHATTACK
		mo.state = S_PLAY_DASH
	elseif P_IsObjectOnGround(mo)
	or player.powers[pw_carry] == CR_ROLLOUT
		if player.speed > 0
			mo.state = S_PLAY_WALK
		else
			mo.state = S_PLAY_STND
		end
	else
		mo.state = S_PLAY_FALL
	end
	
	JES.EndAttack(player)
end

JES.MakeAttackState = function(stateName, spr2, extraTics)
	local stateNum
	SafeFreeslot(stateName)
	stateNum = _G[stateName]
	states[stateNum] = {
		sprite = SPR_PLAY,
		frame = spr2|FF_SPR2ENDSTATE,
		tics = 3,
		action = JES.AttackAction,
		nextstate = stateNum,
		var1 = S_JANA_ATTACKEND,
		var2 = extraTics or 0
	}
end

SafeFreeslot("S_JANA_ATTACKEND")
JES.MakeAttackState("S_JANA_COMBOATTACK1", SPR2_TAL1)
JES.MakeAttackState("S_JANA_COMBOATTACK2", SPR2_TAL2)
JES.MakeAttackState("S_JANA_COMBOATTACK3", SPR2_TAL3, 2)
JES.MakeAttackState("S_JANA_DASHATTACK", SPR2_MLEE)
JES.MakeAttackState("S_JANA_AIRATTACK", SPR2_TWIN, 1)
SafeFreeslot("S_JANA_SABERBEAM", "S_JANA_SABERBEAM_FLAME")

states[S_JANA_ATTACKEND] = {
	sprite = SPR_PLAY,
	tics = 0,
	action = JES.AttackEndAction,
	nextstate = S_PLAY_FALL
}

states[S_JANA_SABERBEAM] = {
	sprite = SPR_JSWD,
	frame = FF_FULLBRIGHT,
	tics = -1,
	var1 = 2,
	var2 = SPR_SSWD
}

states[S_JANA_SABERBEAM_FLAME] = {
	sprite = SPR_NULL,
	tics = 6,
	nextstate = S_LASERFLAME1
}

// -------
// objects
// -------

SafeFreeslot("MT_JANA_SABERHITBOX",
	"MT_JANA_SMALLSABERBEAM", "MT_JANA_LARGESABERBEAM", "MT_JANA_LARGESABERBEAM_HITBOX",
	"MT_JANA_SLICEJUNK")

mobjinfo[MT_JANA_SABERHITBOX] = {
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	height = 48*FRACUNIT,
	radius = (80*FRACUNIT - mobjinfo[MT_PLAYER].radius)/2,
	speed = 100*FRACUNIT, // height while overhead
	painchance = (110*FRACUNIT - mobjinfo[MT_PLAYER].radius)/2, // radius while overhead
	flags = MF_NOCLIPHEIGHT|MF_NOSECTOR|MF_NOGRAVITY
}

mobjinfo[MT_JANA_SMALLSABERBEAM] = {
	spawnstate = S_JANA_SABERBEAM,
	spawnhealth = 1000,
	height = 64*FRACUNIT,
	radius = 24*FRACUNIT,
	speed = 30*FRACUNIT,
	reactiontime = TICRATE/2,
	flags = MF_NOGRAVITY|MF_MISSILE|MF_NOCLIPHEIGHT
}

mobjinfo[MT_JANA_LARGESABERBEAM] = {
	spawnstate = S_JANA_SABERBEAM,
	spawnhealth = 1000,
	height = 96*FRACUNIT,
	radius = 16*FRACUNIT,
	speed = 45*FRACUNIT,
	reactiontime = TICRATE,
	flags = MF_NOGRAVITY|MF_MISSILE|MF_NOCLIPHEIGHT // this has MF_MISSILE to travel up slopes, but it doesn't collide with anything!
}

mobjinfo[MT_JANA_LARGESABERBEAM_HITBOX] = {
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	height = 96*FRACUNIT,
	radius = 64*FRACUNIT,
	flags = MF_NOGRAVITY|MF_MISSILE|MF_NOCLIPHEIGHT|MF_PAPERCOLLISION
}

mobjinfo[MT_JANA_SLICEJUNK] = {
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	deathstate = S_XPLD1,
	deathsound = sfx_cybdth,
	painchance = MT_SMOKE,
	height = 32*FRACUNIT,
	radius = 16*FRACUNIT,
	speed = 10*FRACUNIT
}

// -----------
// definitions
// -----------

JES.SKIN = "jana"

JES.FIRSTATTACKSTATE = S_JANA_COMBOATTACK1
JES.COMBOENDSTATE = S_JANA_COMBOATTACK3
JES.LASTATTACKSTATE = S_JANA_AIRATTACK

JES.COMBOCOYOTETICS = 7

JES.FULLCHARGETIME = 3*TICRATE/2
JES.CHARGETHRESHOLD = JES.FULLCHARGETIME/2
JES.FULLCHARGESOUND = sfx_s3ka8
JES.CHARGESOUND = sfx_s3ka5

JES.PFLAGS_BLACKLIST = PF_SLIDING|PF_BOUNCING|PF_GLIDING|PF_STARTDASH|PF_SPINNING|PF_GAMETYPEOVER // Jana doesn't use half of these, but you just never know

JES.FLAGS = {}
JES.FLAGS.USEDATTACK = 1 << 0 // used to signify that player properties should be reverted after an attack
JES.FLAGS.ATTACKSTASIS = 1 << 1 // player has no control and reduced friction during attack
JES.FLAGS.OVERHEADATTACK = 1 << 2 // attack uses a larger, paper hitbox
JES.FLAGS.CHARGEATTACK = 1 << 3 // this attack was charged
JES.FLAGS.QUEUEATTACK = 1 << 4 // start another attack as soon as possible
JES.FLAGS.FALLING = 1 << 5 // applied whenever momz < 0, to signify that cralwa/monitor bounces should not be possible
JES.FLAGS.SUPERBLOCKED = 1 << 6 // applied when SF_SUPER is removed, to signify SF_SUPER should be restored
JES.FLAGS.ATTACKMASK = JES.FLAGS.USEDATTACK|JES.FLAGS.ATTACKSTASIS|JES.FLAGS.OVERHEADATTACK|JES.FLAGS.CHARGEATTACK|JES.FLAGS.QUEUEATTACK

JES.SLICEFLAGS = {}
JES.SLICEFLAGS.NODONTDRAW = 1 << 0 // do not give the sliced object MF2_DONTDRAW
JES.SLICEFLAGS.NOSMOKE = 1 << 1 // don't make junks spawn smoke

JES.sliceInfo = {
	[MT_BLUECRAWLA] = {
		sprite = SPR_SLCB,
		frames = {A, B}
	},
	[MT_REDCRAWLA] = {
		sprite = SPR_SLCB,
		frames = {C, D}
	},
	[MT_GFZFISH] = {
		sprite = SPR_SLCB,
		frames = {E, F}
	},
	[MT_SNAILER] = {
		sprite = SPR_SLCB,
		frames = {G, H}
	},
	[MT_GOLDBUZZ] = {
		sprite = SPR_SLCB,
		frames = {I, J}
	},
	[MT_REDBUZZ] = {
		sprite = SPR_SLCB,
		frames = {K, L}
	},
	[MT_JETJAW] = {
		sprite = SPR_SLCB,
		frames = {M, N}
	},
	[MT_SPRINGSHELL] = {
		sprite = SPR_SLCB,
		frames = {O, P}
	},
	[MT_YELLOWSHELL] = {
		sprite = SPR_SLCB,
		frames = {Q, R}
	},
	[MT_CRUSHSTACEAN] = {
		sprite = SPR_SLCB,
		frames = {S, T}
	},
	[MT_ROBOHOOD] = {
		sprite = SPR_SLCB,
		frames = {U, V}
	},
	[MT_EGGSHIELD] = {
		sprite = SPR_SLCB,
		frames = {W, X},
		onSlice = function(item, inf, source)
			if inf				
				if inf.player
					S_StartSound(inf, item.info.deathsound)
				elseif valid(source)
					S_StartSound(source, item.info.deathsound)
				end
				
				if (inf.flags & MF_MISSILE)
					P_KillMobj(inf, item, item.target)
				end
			end
			P_RemoveMobj(item)
		end
	},
	[MT_EGGGUARD] = {
		sprite = SPR_SLCB,
		frames = {Y, Z}
	},
	[MT_JETTBOMBER] = {
		sprite = SPR_SLC2,
		frames = {A, B}
	},
	[MT_JETTGUNNER] = {
		sprite = SPR_SLC2,
		frames = {C, D}
	},
	[MT_VULTURE] = {
		sprite = SPR_SLC2,
		frames = {E, F}
	},
	[MT_MINUS] = {
		sprite = SPR_SLC2,
		frames = {G, H}
	},
	[MT_UNIDUS] = {
		sprite = SPR_SLC2,
		frames = {I, J}
	},
	[MT_PENGUINATOR] = {
		sprite = SPR_SLC2,
		frames = {K, L}
	},
	[MT_POPHAT] = {
		sprite = SPR_SLC2,
		frames = {M, N}
	},
	[MT_SPINCUSHION] = {
		sprite = SPR_SLC2,
		frames = {O, P}
	},
	[MT_CACOLANTERN] = {
		sprite = SPR_SLC2,
		frames = {Q, R},
		onJunkDeath = function(junk)
			local stateNums = {S_CACO_DIE_GIB1, S_CACO_DIE_GIB2, S_CACO_DIE_SHATTER}
			local thok = P_SpawnMobjFromMobj(junk, 0, 0, 0, MT_THOK)
			thok.flags2 = $ | MF2_DONTDRAW
			thok.tics = 2*TICRATE
			for i, num in ipairs(stateNums)
				local state = states[num]
				state.action(thok, state.var1, state.var2)
			end
			P_RemoveMobj(junk)
			return true
		end
	},
	[MT_BANPYURA] = {
		sprite = SPR_SLC2,
		frames = {S, T}
	},
	[MT_HANGSTER] = {
		sprite = SPR_SLC2,
		frames = {U, V}
	},
	[MT_PTERABYTE] = {
		sprite = SPR_SLC2,
		frames = {W, X}
	},
	[MT_PYREFLY] = {
		sprite = SPR_SLC2,
		frames = {Y, Z}
	},
	monitor = {
		sprite = SPR_SLC3,
		frames = {A, B},
		flags = JES.SLICEFLAGS.NODONTDRAW
	},
	funcs = {
		// tip: you could add your own functions here with table.insert(JanaEnergySaber.sliceInfo.funcs, 1, yourFunction) ;D
		function(item)
			if item.flags & (MF_MONITOR|MF_GRENADEBOUNCE) == MF_MONITOR
				return JES.sliceInfo.monitor
			end
		end
	}
}

// -----
// debug
// -----

JES.DEBUGPREFIX = "\141Jana\136EnergySaber\141:\128 "

JES.debug = CV_RegisterVar(
{
	name = "jana_debug",
	PossibleValue = CV_OnOff,
	defaultvalue = "Off"
})

JES.DebugPrint = function(str)
	if JES.debug.value
		print(JES.DEBUGPREFIX .. str)
	end
end

JES.DebugHUD = function(v, player)
	local jes = player.janaEnergySaber
	
	if not jes
	or not JES.debug.value
		return
	end
	
	local STEP = 4
	local x = 16
	local y = 56
	local flags = V_SNAPTOLEFT|V_SNAPTOTOP
	
	local function DrawString(str, extraFlags)
		if str ~= nil
			v.drawString(x, y, str, flags|(extraFlags or 0), "small")
		end
		y = $ + STEP
	end
	
	DrawString("Flags:")
	
	for flag, value in pairs(JES.FLAGS)
		if flag:find("MASK")
			continue
		end
		
		DrawString(flag, (jes.flags & value) and V_GREENMAP or V_REDMAP)
	end
	
	DrawString()
	DrawString("Combo coyote time:")
	DrawString(jes.comboCoyoteTics, jes.comboCoyoteTics > 0 and V_GREENMAP or V_REDMAP)
	
	DrawString()
	DrawString("Charge time:")
	DrawString(jes.chargeTime,
		jes.chargeTime < JES.CHARGETHRESHOLD and V_REDMAP
		or jes.chargeTime < JES.FULLCHARGETIME and V_ORANGEMAP
		or V_GREENMAP)
end

// ---------
// functions
// ---------

// main function, to be hooked in ThinkFrame
JES.Main = function()
	if JES.AngleAttacks then
		JES.AngleAttacks.CheckConfig()
	end
	for player in players.iterate
		JES.UpdatePlayer(player)
	end
end

// main player processor
JES.UpdatePlayer = function(player)
	if not JES.CheckStatus(player)
		return
	end
	JES.UpdateButtons(player)
	JES.CheckSkinFlags(player)
	JES.CheckFalling(player)
	JES.HandleSaber(player)
end

// returns true if this player is alive and Jana, false if not. takes care of creating or removing the table of variables as appropriate.
JES.CheckStatus = function(player)
	if player.skin ~= JES.GetSkinNumber()
	or player.playerstate ~= PST_LIVE
	or not player.mo
		JES.ResetPlayerTable(player)
		return false
	end
	if player.janaEnergySaber == nil
		player.janaEnergySaber = JES.CreateTable()
	end
	return true
end

// counts how long the player has held spin for
JES.UpdateButtons = function(player)
	local jes = player.janaEnergySaber
	
	if player.cmd.buttons & BT_USE
	and not player.exiting
	and not player.powers[pw_nocontrol]
	and not (player.pflags & PF_STASIS)
		jes.spinHeld = $ + 1
	else
		jes.spinHeld = 0
	end
end

// certain skin flags should be added or removed under certain conditions
JES.CheckSkinFlags = function(player)
	local jes = player.janaEnergySaber
	
	// SF_NOSHIELDABILITY should be removed if Jana is in midair and not jumping, so she can use whirlwind jump
	if not (player.pflags & PF_JUMPED)
	and not P_IsObjectOnGround(player.mo)
		player.charflags = $ & ~SF_NOSHIELDABILITY
	else
		player.charflags = $ | (skins[player.skin].flags & SF_NOSHIELDABILITY)
	end
	
	// SF_SUPER should be restored if it was removed earlier
	if jes.flags & JES.FLAGS.SUPERBLOCKED
		jes.flags = $ & ~JES.FLAGS.SUPERBLOCKED
		player.charflags = $ | (skins[player.skin].flags & SF_SUPER)
	end
end

// flags Jana as falling so she can ignore crawla/monitor bounces while using her saber
JES.CheckFalling = function(player)
	local mo = player.mo
	local jes = player.janaEnergySaber
	
	if P_IsObjectOnGround(mo)
	or P_MobjFlip(mo)*mo.momz > 0
		jes.flags = $ & ~JES.FLAGS.FALLING
	else
		jes.flags = $ | JES.FLAGS.FALLING
	end
end

// handles Jana's attacks
JES.HandleSaber = function(player)
	local mo = player.mo
	local jes = player.janaEnergySaber
	
	// charge an attack if spin is held
	if not JES.CanChargeAttack(player)
		jes.chargeTime = 0
	elseif jes.spinHeld
		if jes.chargeTime < JES.FULLCHARGETIME
			jes.chargeTime = $ + 1
			if jes.chargeTime >= JES.FULLCHARGETIME
				local ghost = P_SpawnGhostMobj(mo)
				ghost.scalespeed = FixedMul($, ghost.scale)
				ghost.destscale = $ << 1
				ghost.color = SKINCOLOR_WHITE
				ghost.colorized = true
				ghost.blendmode = AST_ADD
				ghost.renderflags = $ | RF_FULLBRIGHT
				S_StartSound(mo, JES.FULLCHARGESOUND)
				JES.SpawnChargeEffect(player)
			end
		end
		
		if jes.chargeTime >= JES.CHARGETHRESHOLD
			if jes.chargeTime == JES.CHARGETHRESHOLD
				S_StartSound(mo, JES.CHARGESOUND)
			end
			JES.SpawnParticle(player, jes.chargeTime * mo.scale / JES.FULLCHARGETIME)
		end
	elseif jes.chargeTime >= JES.CHARGETHRESHOLD
	and JES.CanStartAttack(player)
		jes.flags = $ | JES.FLAGS.QUEUEATTACK
	else
		if jes.chargeTime >= JES.CHARGETHRESHOLD
			S_StopSoundByID(mo, JES.CHARGESOUND)
		end
		jes.chargeTime = 0
	end
	
	// start an attack if spin is pressed or a combo attack is queued
	if jes.spinHeld == 1
	or jes.flags & JES.FLAGS.QUEUEATTACK
		if JES.CanStartAttack(player)
			JES.StartAttack(player)
		elseif mo.state >= JES.FIRSTATTACKSTATE
		and mo.state < JES.COMBOENDSTATE
			jes.flags = $ | JES.FLAGS.QUEUEATTACK
		else
			jes.flags = $ & ~JES.FLAGS.QUEUEATTACK
		end
	end
	
	// while attacking,
	if JES.InAttackState(player)
		// reduce control if needed
		if jes.flags & JES.FLAGS.ATTACKSTASIS
			player.pflags = $ | PF_JUMPSTASIS
			player.thrustfactor = 0
			mo.friction = $ + (FRACUNIT - $)/2
		end
		
		// force drawangle to the camera angle if necessary
		if JES.UsingStandardMode(player)
		and not JES.AttackFacesDrawangle(player)
			player.drawangle = mo.angle
		end
		
		// move the saber hitbox
		JES.MoveSaberHitbox(player)
		return
	end
	
	// count down coyote time
	if jes.comboCoyoteTics > 0
		if JES.CanStartAttack(player)
			jes.comboCoyoteTics = $ - 1
		else
			jes.comboCoyoteTics = 0
		end
	end
		
	// undo player modifications if needed
	JES.EndAttack(player)
end

// returns whether the player's current state is one of Jana's attack states
JES.InAttackState = function(player)
	local mo = player.mo
	
	return mo.state >= JES.FIRSTATTACKSTATE
	and mo.state <= JES.LASTATTACKSTATE
end

// returns whether the player's current state is one of the combo (grounded) attack states
JES.InComboState = function(player)
	local mo = player.mo
	
	return mo.state >= JES.FIRSTATTACKSTATE
	and mo.state <= JES.COMBOENDSTATE
end

// returns whether the player is in a super transformation state
JES.InTransformState = function(player)
	local mo = player.mo
	
	return mo.state >= S_PLAY_SUPER_TRANS1
	and mo.state <= S_PLAY_SUPER_TRANS6
end

// returns whether this player can start an attack
JES.CanStartAttack = function(player)
	local mo = player.mo
	local jana = player.jana
	
	if JES.InAttackState(player) // don't attack if attacking,
	and not JES.CanComboAttack(player) // unless you can cancel your attack!
	or JES.InTransformState(player)
	or P_PlayerInPain(player)
	or player.powers[pw_carry] ~= CR_NONE and player.powers[pw_carry] ~= CR_ROLLOUT
	or player.pflags & (JES.PFLAGS_BLACKLIST)
	or jana and (jana.diving or jana.walltouch > 0 or jana.rockin or jana.drippin)
		return false
	end
	return true
end

// returns whether this player can charge an attack by holding spin
JES.CanChargeAttack = function(player)
	local mo = player.mo
	local jana = player.jana
	
	if JES.InTransformState(player)
	or P_PlayerInPain(player)
	or player.powers[pw_carry] ~= CR_NONE and player.powers[pw_carry] ~= CR_ROLLOUT
	or player.pflags & PF_GAMETYPEOVER
	or jana and (jana.rockin or jana.drippin)
		return false
	end
	return true
end

// returns whether or not this attack can be cancelled into another attack
JES.CanComboAttack = function(player)
	local mo = player.mo
	
	return mo.state >= JES.FIRSTATTACKSTATE
		and mo.state < JES.COMBOENDSTATE
		and (mo.frame & FF_FRAMEMASK) == skins[player.skin].sprites[mo.sprite2].numframes - 1
end

// Starts an attack based on the current circumstances of the player
JES.StartAttack = function(player)
	local mo = player.mo
	local jana = player.jana
	local jes = player.janaEnergySaber
	local sound = sfx_jaatk1 + jes.soundToggle
	local coyoteTics = jes.comboCoyoteTics
	local chargeAttack = jes.chargeTime >= JES.FULLCHARGETIME and JES.FLAGS.CHARGEATTACK or 0
	
	JES.EndAttack(player)
	jes.flags = $ & ~JES.FLAGS.QUEUEATTACK | JES.FLAGS.USEDATTACK | chargeAttack
	
	// choose attack and sound
	if P_IsObjectOnGround(mo)
	or player.powers[pw_carry] == CR_ROLLOUT
		if jana and jana.dashing // dash attack
			mo.state = S_JANA_DASHATTACK
		else // combo attack
			if chargeAttack
				mo.state = JES.COMBOENDSTATE
			elseif JES.CanComboAttack(player)
			or jes.comboCoyoteTics > 0
				mo.state = jes.lastAttackState + 1
			else
				mo.state = S_JANA_COMBOATTACK1
			end
			
			if mo.state == JES.COMBOENDSTATE // attack 3 is overhead!
				jes.flags = $ | JES.FLAGS.OVERHEADATTACK
			end
			
			jes.flags = $ | JES.FLAGS.ATTACKSTASIS
			sound = sfx_jaatk1 + (mo.state - S_JANA_COMBOATTACK1)
		end
	else // air attack
		mo.state = S_JANA_AIRATTACK
		jes.flags = $ | JES.FLAGS.OVERHEADATTACK
	end
	jes.lastAttackState = mo.state
	
	// set coyote tics
	if mo.state >= JES.FIRSTATTACKSTATE
	and mo.state < JES.COMBOENDSTATE
		jes.comboCoyoteTics = JES.COMBOCOYOTETICS
	else
		jes.comboCoyoteTics = 0
	end
	
	// no skidding
	player.charflags = $ | SF_NOSKID
	player.skidtime = 0
	
	// cancel dive landing
	if jana
		jana.controllock = 0
	end
	
	// alternate between first and second attack sounds for dash/air attacks
	if sound == sfx_jaatk1 + jes.soundToggle
		jes.soundToggle = $ ^^ 1
	end
	S_StartSound(mo, sound)
end

// reverts modifications made to player after an attack
JES.EndAttack = function(player)
	local jes = player.janaEnergySaber
	
	if not (jes.flags & JES.FLAGS.USEDATTACK) // nothing to revert
		return
	end
	
	if jes.flags & JES.FLAGS.ATTACKSTASIS // we need to reset our thrustfactor
		player.thrustfactor = skins[player.skin].thrustfactor
	end
	player.charflags = $ & ~SF_NOSKID | (skins[player.skin].flags & SF_NOSKID)
	jes.flags = $ & ~JES.FLAGS.ATTACKMASK
end

// creates and moves the saber's hitbox
JES.MoveSaberHitbox = function(player)
	local mo = player.mo
	local jes = player.janaEnergySaber
	local hitbox = jes.hitbox
	local targetX, targetY, targetZ
	
	// create a hitbox if one does not exist
	if not valid(hitbox)
		hitbox = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_JANA_SABERHITBOX)
		hitbox.target = mo
		jes.hitbox = hitbox
	end
	
	// ensure hitbox scale matches the player's
	if hitbox.scale ~= mo.scale
		hitbox.scale = mo.scale
	end
	
	// modify the hitbox based on whether it's an overhead attack or not
	if jes.flags & JES.FLAGS.OVERHEADATTACK
		if not (hitbox.flags & MF_PAPERCOLLISION)
			hitbox.flags = $ | MF_PAPERCOLLISION
			hitbox.height = FixedMul(hitbox.info.speed, hitbox.scale)
			hitbox.radius = FixedMul(hitbox.info.painchance, hitbox.scale)
		end
	else
		if (hitbox.flags & MF_PAPERCOLLISION)
			hitbox.flags = $ & ~MF_PAPERCOLLISION
			hitbox.radius = FixedMul(hitbox.info.radius, hitbox.scale)
			hitbox.height = FixedMul(hitbox.info.height, hitbox.scale)
		end
	end
	
	// move the hitbox to the correct position
	hitbox.angle = player.drawangle
	targetX = mo.x + P_ReturnThrustX(mo, hitbox.angle, mo.radius + hitbox.radius)
	targetY = mo.y + P_ReturnThrustY(mo, hitbox.angle, mo.radius + hitbox.radius)
	if mo.eflags & MFE_VERTICALFLIP
		targetZ = mo.z + mo.height - hitbox.height
	else
		targetZ = mo.z
	end
	
	if P_TryMove(hitbox, targetX, targetY, true) // smoothly move the hitbox so it catches things in between positions
		hitbox.z = targetZ
	elseif hitbox.valid // fuck you, you're going there anyway
		P_MoveOrigin(hitbox, targetX, targetY, targetZ)
	end
end

// spawns a projectile appropriate to the player's current combo attack
JES.SpawnSaberBeam = function(player)
	local mo = player.mo
	local jes = player.janaEnergySaber
	local mobjType
	local angle
	
	if mo.state == JES.COMBOENDSTATE
	or jes and jes.flags & JES.FLAGS.CHARGEATTACK
		mobjType = MT_JANA_LARGESABERBEAM
	else
		mobjType = MT_JANA_SMALLSABERBEAM
	end
	
	if JES.AttackFacesDrawangle(player)
		angle = player.drawangle
	else
		angle = mo.angle
	end
	
	local radius = FixedDiv(mo.radius, mo.scale) + mobjinfo[MT_JANA_SABERHITBOX].radius
	local x = P_ReturnThrustX(mo, angle, radius)
	local y = P_ReturnThrustY(mo, angle, radius)
	
	local beam = P_SpawnMobjFromMobj(mo, x, y, 0, mobjType)
	
	beam.target = mo
	beam.scalespeed = FixedMul($, beam.scale)
	beam.angle = angle
	beam.momx = mo.momx
	beam.momy = mo.momy
	P_Thrust(beam, beam.angle, FixedMul(beam.info.speed, beam.scale))
	
	JES.EnsureMinSpeed(beam)
	
	// set fuse
	if player.powers[pw_super]
		if mobjType == MT_JANA_LARGESABERBEAM
			beam.flags2 = $ | MF2_JUSTATTACKED // flag to spawn flames
		end
		beam.destscale = 3*$/2
		beam.fuse = 2*beam.info.reactiontime
		beam.color = SKINCOLOR_PURPLE
	else
		beam.fuse = beam.info.reactiontime
		beam.color = SKINCOLOR_SKY
	end
	
	// set sprite and frame
	if beam.state == S_JANA_SABERBEAM
		if player.powers[pw_super]
			beam.sprite = states[beam.state].var2
		end
		if beam.type == MT_JANA_LARGESABERBEAM
			beam.frame = $ + states[beam.state].var1 * 2
		else
			beam.frame = $ + states[beam.state].var1 * ((mo.state - JES.FIRSTATTACKSTATE) & 1)
		end
	end
	
	// spawn hitboxes for large beam
	if mobjType == MT_JANA_LARGESABERBEAM
		for i = 0, 1
			local box = P_SpawnMobjFromMobj(beam, 0, 0, 0, MT_JANA_LARGESABERBEAM_HITBOX)
			box.target = mo
			box.tracer = beam
			box.movedir = ANGLE_90 + i*ANGLE_180
			JES.SaberBeamHitboxThinker(box)
		end
	end
end

// ensures an object is traveling at least as fast as its speed in the direction of its angle
JES.EnsureMinSpeed = function(beam)
	local targetSpeed = FixedMul(beam.info.speed, beam.scale)
	local angle = R_PointToAngle2(0, 0, beam.momx, beam.momy) - beam.angle
	local speed = FixedHypot(beam.momx, beam.momy)
	local speedAtAngle = FixedMul(speed, cos(angle))
	
	if speedAtAngle < targetSpeed
		P_Thrust(beam, beam.angle, targetSpeed - speedAtAngle)
	end
end

// spawns an energy ball at a random place around @player
JES.SpawnParticle = function(player, scale)
	local mo = player.mo
	local flip = P_MobjFlip(mo)
	local radius = FixedDiv(mo.radius, mo.scale) << 1
	local height = FixedDiv(mo.height, mo.scale)
	local ball = P_SpawnMobjFromMobj(mo,
		FixedMul(P_RandomFixed(), radius << 1) - radius,
		FixedMul(P_RandomFixed(), radius << 1) - radius,
		FixedMul(P_RandomFixed(), height), MT_JANA_ENERGYBALL)
	ball.scalespeed = FixedMul($, mo.scale)
	ball.scale = scale
	ball.destscale = 0
	P_SetObjectMomZ(ball, 8*FRACUNIT)
	ball.momz = $ + flip*max(0, flip*JES.GetMobjZMovement(mo))
	ball.fuse = TICRATE/3
	ball.target = mo
end

// spawns the little charge effect if it doesn't exist already
JES.SpawnChargeEffect = function(player)
	if not JES.CheckStatus(player)
		return
	end
	
	local mo = player.mo
	local jes = player.janaEnergySaber
	
	if not valid(jes.chargeEffect)
		jes.chargeEffect = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_JANA_CHARGEFULL)
		jes.chargeEffect.target = mo
	end
end

// creates and returns a table with all the Jana attack variables the player needs
JES.CreateTable = function()
	return {
		spinHeld = 0,
		flags = 0,
		lastAttackState = 0,
		soundToggle = 0,
		comboCoyoteTics = 0,
		chargeTime = 0
	}
end

// resets all player values and removes the variable table
JES.ResetPlayerTable = function(player)
	if player.janaEnergySaber == nil // nothing to do
		return
	end
	
	// reset thrustfactor if in attack stasis
	if player.skin == JES.GetSkinNumber()
		JES.EndAttack(player)
	end
	
	// remove the table
	player.janaEnergySaber = nil
end

// returns Jana's skin number, because comparing numbers is faster than comparing strings :D
JES.GetSkinNumber = function()
	if JES.skinNumber == nil
		JES.skinNumber = #skins[JES.SKIN]
	end
	return JES.skinNumber
end

// returns whether this player is using the Standard control style (directionchar)
JES.UsingStandardMode = function(player)
	return (player.pflags & (PF_ANALOGMODE|PF_DIRECTIONCHAR)) == PF_DIRECTIONCHAR
end

// returns whether or not the player's attack should face the player's drawangle
JES.AttackFacesDrawangle = function(player)
	local mo = player.mo
	local value = player.jana_angleattacks
	
	if value == nil
		value = JES.AngleAttacks.DEFAULT
	end
	
	if JES.InComboState(player)
		return value & JES.AngleAttacks.GROUND
	end
	
	if mo.state == S_JANA_DASHATTACK
		return value & JES.AngleAttacks.DASH
	end
	
	if mo.state == S_JANA_AIRATTACK
		return value & JES.AngleAttacks.AIR
	end
	
	return true // always use drawangle if not attacking
end

// returns whether or not this point is considered in front of Jana
JES.PointInFront = function(player, x, y)
	local mo = player.mo
	local jes = player.janaEnergySaber
	local playerX = mo.x - mo.momx - P_ReturnThrustX(mo, player.drawangle, mo.radius >> 1)
	local playerY = mo.y - mo.momy - P_ReturnThrustY(mo, player.drawangle, mo.radius >> 1)
	local itemDirection = R_PointToAngle2(playerX, playerY, x, y)
	local angleSpan = (jes.flags & JES.FLAGS.OVERHEADATTACK) and ANGLE_45 or ANGLE_90
	
	return abs(player.drawangle - itemDirection) < angleSpan
end

// returns the closest point on a line, taking its vertices into account
JES.ClosestPointOnLine = function(line, x, y)
	local v1 = line.v1
	local v2 = line.v2
	local minX = min(v1.x, v2.x)
	local maxX = max(v1.x, v2.x)
	local minY = min(v1.y, v2.y)
	local maxY = max(v1.y, v2.y)
	
	x, y = P_ClosestPointOnLine(x, y, line)
	
	x = min(max($, minX), maxX)
	y = min(max($, minY), maxY)
	
	return x, y
end

// busts any breakable blocks within @mo's height range in the sectors around @line
JES.MobjBustFOFsAroundLine = function(mo, line)
	for i, sector in ipairs({line.frontsector, line.backsector})
		for fof in sector.ffloors()
			if fof.flags & (FF_EXISTS|FF_BUSTUP) ~= (FF_EXISTS|FF_BUSTUP)
				continue
			end
			
			local top
			local bottom
			
			if fof.t_slope
				top = P_GetZAt(fof.t_slope, x, y)
			else
				top = fof.topheight
			end
			
			if fof.b_slope
				bottom = P_GetZAt(fof.b_slope, x, y)
			else
				bottom = fof.bottomheight
			end
			
			if mo.z < top
			and mo.z + mo.height > bottom
				JES.MobjBustFOF(mo, fof)
			end
		end
	end
end

// makes @mobj crumble @fof
JES.MobjBustFOF = function(mo, fof)
	local line = fof.master
	
	// bust the FOF
	EV_CrumbleChain(fof)
	
	// run a linedef executor?
	if line.flags & ML_EFFECT5
		P_LinedefExecute(P_AproxDistance(line.dx, line.dy) >> FRACBITS, mo, fof.target)
	end
end

// returns the "z movement" of @mo, using momz in the air or the slope pitch on the ground
JES.GetMobjZMovement = function(mo)
	if P_IsObjectOnGround(mo)
		local slope = mo.standingslope
		
		if slope
			local angDiff = R_PointToAngle2(0, 0, mo.momx, mo.momy) - slope.xydirection
			local speed = FixedHypot(mo.momx, mo.momy)
			
			return FixedMul(FixedMul(speed, cos(angDiff)), sin(slope.zangle))
		end
		
		return 0
	end
	
	return mo.momz
end

// returns whether @mo is touching the ground (works in goop too)
JES.MobjTouchingGround = function(mo)
	if mo.eflags & MFE_VERTICALFLIP
		return mo.z + mo.height >= mo.ceilingz
	end
	return mo.z <= mo.floorz
end

// returns whether @mo is an attack from Jana and that @item is being attacked by @mo
JES.MobjCollisionIsAttack = function(item, mo)
	if not item.valid
	or not mo.valid
	or mo.z > item.z + item.height
	or item.z > mo.z + mo.height
		return false
	end
	
	local player = mo.player
	local validAttacker = mo.type == MT_JANA_SABERHITBOX
		or mo.type == MT_JANA_SMALLSABERBEAM
		or mo.type == MT_JANA_LARGESABERBEAM_HITBOX
		or player
		and JES.CheckStatus(player)
		and JES.InAttackState(player)
		and JES.PointInFront(player, item.x, item.y)
	
	return validAttacker
end

// removes SF_SUPER if not holding jump
JES.SpinSpecial = function(player)
	if not JES.CheckStatus(player)
		return
	end
	
	local jes = player.janaEnergySaber
	
	if not (player.cmd.buttons & BT_JUMP)
	and not player.powers[pw_super]
		player.charflags = $ & ~SF_SUPER
		jes.flags = $ | JES.FLAGS.SUPERBLOCKED
	end
end

// damages objects in front of Jana, to be hooked with PlayerCanDamage
JES.PlayerCanDamage = function(player, item)
	if not JES.CheckStatus(player)
	or not JES.InAttackState(player)
	or not item.valid
		return
	end
	
	return JES.PointInFront(player, item.x, item.y)
end

// resets variables on respawn
JES.PlayerSpawn = JES.ResetPlayerTable

// breaks walls in front of Jana, to be hooked with MobjLineCollide
JES.PlayerMobjLineCollide = function(mo, line)
	local player = mo.valid and mo.player
	
	if not player
	or not JES.CheckStatus(player)
	or not JES.InAttackState(player)
		return
	end
	
	local x, y = JES.ClosestPointOnLine(line, mo.x, mo.y)
	
	if not JES.PointInFront(player, x, y)
		return
	end
	
	JES.MobjBustFOFsAroundLine(mo, line)
end

// --------------------
// MobjDamage functions
// --------------------

JES.MobjDamage = function(mo, inf, source, damage, damageType)
	JES.CheckObjectBounce(mo, inf)
end

// prevents Jana from crawla/monitor bouncing
JES.CheckObjectBounce = function(item, mo)
	local player = valid(mo) and mo.player
	
	if not player
	or not JES.CheckStatus(player)
	or not JES.InAttackState(player) // don't worry about bounces that happen outside of attacks
		return
	end
	
	local jes = player.janaEnergySaber
	
	// revert vertical bounces
	if jes.flags & JES.FLAGS.FALLING
	and P_MobjFlip(mo)*mo.momz > 0
		mo.momz = -$
	end
	
	// revert horizontal bounces
	if item.valid and item.info.spawnhealth > 1
		mo.momx = -$
		mo.momy = -$
	end
end

// -------------------
// MobjDeath functions
// -------------------

JES.MobjDeath = function(mo, inf, source, damageType)
	JES.SliceObject(mo, inf, source)
end

// if an object is sliceable, slices it!
JES.SliceObject = function(item, inf, source)
	if not item.valid
	or not valid(inf) // if you weren't damaged by an object, Jana *definitely* didn't slice you :p
		return
	end
	
	local isSliced = inf.player and JES.CheckStatus(inf.player) and JES.InAttackState(inf.player)
		or inf.type == MT_JANA_SABERHITBOX
		or inf.type == MT_JANA_SMALLSABERBEAM
		or inf.type == MT_JANA_LARGESABERBEAM // this should not ever damage objects, but account for it anyway
		or inf.type == MT_JANA_LARGESABERBEAM_HITBOX
	
	if not isSliced
		return
	end
	
	local sliceInfo = JES.GetSliceInfo(item)
	
	if not sliceInfo
		return
	end
	
	local flags = sliceInfo.flags or 0
	local numFrames = #sliceInfo.frames
	local dir = 2*P_RandomKey(2) - 1 // 1 or -1
	local turn = FixedAngle(360*FRACUNIT/numFrames)
	local roll = dir*ANG20
	local angle = dir*ANGLE_90
	
	if inf.player
		angle = $ + inf.player.drawangle
	else
		angle = $ + inf.angle
	end
	
	if not (flags & JES.SLICEFLAGS.NODONTDRAW)
		item.flags2 = $ | MF2_DONTDRAW
	end
	
	for i = 1, numFrames
		local junk = P_SpawnMobjFromMobj(item, 0, 0, FRACUNIT, MT_JANA_SLICEJUNK)
		junk.angle = angle
		junk.sprite = sliceInfo.sprite
		junk.frame = sliceInfo.frames[i] | (item.frame & ~FF_FRAMEMASK)
		junk.movedir = roll
		junk.extravalue1 = item.type
		junk.extravalue2 = flags
		junk.target = source
		junk.tracer = item
		P_InstaThrust(junk, junk.angle, FixedMul(junk.info.speed, junk.scale) >> 1)
		P_SetObjectMomZ(junk, junk.info.speed)
		
		angle = $ + turn
		roll = -$
	end
	
	if sliceInfo.onSlice
		sliceInfo.onSlice(item, inf, source)
	end
end

JES.GetSliceInfo = function(item)
	local info
	for i, func in ipairs(JES.sliceInfo.funcs)
		info = func(item)
		if info
			return info
		end
	end
	return JES.sliceInfo[item.type]
end

// ---------------------------
// MT_JANA_SLICEJUNK functions
// ---------------------------

JES.SliceJunkThinker = function(junk)
	if not junk.valid
	or junk.health <= 0
		return
	end
	
	junk.rollangle = $ + junk.movedir
	
	if not (junk.extravalue2 & JES.SLICEFLAGS.NOSMOKE)
	and leveltime % 4 == 0
		A_FlameParticle(junk)
	end
	
	// used to be P_IsObjectOnGround, but I need it to work in goo too
	if JES.MobjTouchingGround(junk)
		JES.KillSliceJunk(junk)
	end
end

JES.SliceJunkMoveBlocked = function(junk)
	if not junk.valid
	or junk.health <= 0
		return
	end
	
	JES.KillSliceJunk(junk)
end

JES.KillSliceJunk = function(junk)
	local tracer = junk.tracer
	local sliceInfo = valid(tracer) and JES.GetSliceInfo(tracer)
		or JES.sliceInfo[junk.extravalue1]
	
	junk.rollangle = 0
	junk.momx = 0
	junk.momy = 0
	junk.momz = 0
	
	if sliceInfo
	and sliceInfo.onJunkDeath
	and sliceInfo.onJunkDeath(junk)
		return
	else
		P_KillMobj(junk)
	end
end

// -----------------------------
// MT_JANA_SABERHITBOX functions
// -----------------------------

JES.SaberHitboxThinker = function(hitbox)
	if not hitbox.valid
		return
	end
	
	local mo = hitbox.target
	local player = valid(mo) and mo.player
	
	// remove hitbox if player is not an attacking Jana
	if not player
	or not JES.CheckStatus(player)
	or not JES.InAttackState(player)
		P_RemoveMobj(hitbox)
		return
	end
end

JES.SaberHitboxCollide = function(hitbox, item)
	if not hitbox.valid
	or not item.valid
	or hitbox.z > item.z + item.height
	or item.z > hitbox.z + hitbox.height
		return
	end
	
	local mo = hitbox.target
	
	if not (item.flags & MF_SHOOTABLE) // only damage damageable objects
	or item.health <= 0 // you can't damage dead items, silly
	or item.player // I'll let this damage enemy players later
	//or item == mo // don't collide with your owner, duh
		return
	end
	
	// check that your owner is valid before you destroy anything for them
	local player = valid(mo) and mo.player
	
	if not player
	or not JES.CheckStatus(player)
	or not JES.InAttackState(player)
		P_RemoveMobj(hitbox)
		return
	end
	
	// ok, should be safe to attack!
	P_DamageMobj(item, hitbox, mo)
end

JES.SaberHitboxLineCollide = function(hitbox, line)
	if not hitbox.valid
		return
	end
	
	JES.MobjBustFOFsAroundLine(hitbox, line)
end

// --------------------
// Saber beam functions
// --------------------

JES.SaberBeamThinker = function(beam)
	if not beam.valid
		return
	end
	
	// spawn flames if super
	if beam.flags2 & MF2_JUSTATTACKED
	and beam.fuse & 1
		local doFlames = false
		
		if beam.eflags & MFE_VERTICALFLIP
			if beam.z + beam.height > beam.ceilingz - 24*beam.scale
				doFlames = true
			end
		else
			if beam.z < beam.floorz + 24*beam.scale
				doFlames = true
			end
		end
		
		if doFlames
			local flame = P_SpawnMobjFromMobj(beam, 0, 0, 0, MT_SPINFIRE)
			flame.target = beam.target
			flame.scale = $ >> 1
			if flame.eflags & MFE_VERTICALFLIP
				flame.z = flame.ceilingz - flame.height
			else
				flame.z = flame.floorz
			end
			flame.state = S_JANA_SABERBEAM_FLAME
			flame.angle = R_PointToAngle2(0, 0, beam.momx, beam.momy)
			flame.color = beam.color
			flame.colorized = true
		end
	end
	
	// spawn ghost trails
	if beam.fuse % 3 == 2
		P_SpawnGhostMobj(beam)
	end
	
	// shrink when nearing the end of lifespan
	if beam.fuse == 8
		beam.destscale = 0
		beam.scalespeed = (beam.scale - beam.destscale)/beam.fuse
	end
	
	// frame flicker
	beam.frame = $ ^^ 1
end

JES.SaberBeamLineCollide = function(beam, line)
	if not beam.valid
		return
	end
	
	JES.MobjBustFOFsAroundLine(beam, line)
	
	// don't destroy the missile if it barely grazed a wall
	local x, y = JES.ClosestPointOnLine(line, beam.x, beam.y)
	
	if R_PointToDist2(beam.x, beam.y, x, y) > FixedMul(beam.info.speed, beam.scale)
		return false
	end
end

JES.SaberBeamHitboxThinker = function(box)
	if not box.valid
		return
	end
	
	local tracer = box.tracer
	
	if not valid(tracer)
		P_RemoveMobj(box)
		return
	end
	
	local x = tracer.x + P_ReturnThrustX(box, tracer.angle + box.movedir, tracer.radius)
	local y = tracer.y + P_ReturnThrustY(box, tracer.angle + box.movedir, tracer.radius)
	local z
	
	if tracer.eflags & MFE_VERTICALFLIP
		z = tracer.z + tracer.height - box.height
	else
		z = tracer.z
	end
	
	box.angle = tracer.angle
	box.destscale = tracer.destscale
	box.scalespeed = tracer.scalespeed
	
	if P_TryMove(box, x, y, true)
		box.z = z
	elseif box.valid // check validity in case it was removed during TryMove
		P_MoveOrigin(box, x, y, z)
	end
end

JES.SaberBeamHitboxRemoved = function(box)
	local tracer = box.tracer
	
	if not valid(tracer) // don't respawn anything if the beam is also dead
		return
	end
	
	local newBox = P_SpawnMobjFromMobj(box, 0, 0, 0, box.type)
	newBox.movedir = box.movedir
	newBox.target = box.target
	newBox.tracer = tracer
	newBox.angle = tracer.angle
	newBox.destscale = tracer.destscale
	newBox.scalespeed = tracer.scalespeed
end

JES.SaberBeamHitboxCollide = function(box, item)
	if not item.valid
	or not box.valid
		return
	end
	
	if item.flags & (MF_SHOOTABLE|MF_SOLID) == MF_SOLID
	and R_PointToDist2(box.x, box.y, item.x, item.y) > FixedMul(box.info.speed, box.scale) + item.radius
		return false
	end
end

JES.SaberBeamHitboxLineCollide = function()
	return false // never collide with walls, leave that to the beam
end

JES.LargeSaberBeamCollide = function()
	return false // never collide with objects, leave that to the hitboxes
end

JES.SmallSaberBeamCollide = JES.SaberBeamHitboxCollide

// -------------------------
// Spike collision functions
// -------------------------

JES.KillSpike = function(item, mo)
	local spikesToKill = {}
	
	if item.flags & MF_SOLID
		S_StartSound(mo, item.info.deathsound)
	end
	
	for obj in item.subsector.sector.thinglist()
		if (obj.type == item.type and obj.health > 0 and obj.flags & MF_SOLID
		and (obj == item or P_AproxDistance(P_AproxDistance(obj.x - item.x, obj.y - item.y), obj.z - item.z) < 56*item.scale))
			table.insert(spikesToKill, obj)
		end
	end
	
	for i, obj in ipairs(spikesToKill)
		P_KillMobj(obj, mo, mo)
	end
end

JES.SpikeCollide = function(item, mo)
	if JES.MobjCollisionIsAttack(item, mo)
		JES.KillSpike(item, mo)
	end
end

// -------------------------------------------------------------
// Egg Guard shield collision functions (LOL wtf is this object)
// -------------------------------------------------------------

JES.EggShieldCollide = function(item, mo)
	if JES.MobjCollisionIsAttack(item, mo)
	and item.health > 0
		P_KillMobj(item, mo, mo.target)
	end
end

// hook everything!

addHook("ThinkFrame", JES.Main)

addHook("SpinSpecial", JES.SpinSpecial)
addHook("PlayerCanDamage", JES.PlayerCanDamage)
addHook("PlayerSpawn", JES.PlayerSpawn)
addHook("MobjLineCollide", JES.PlayerMobjLineCollide, MT_PLAYER)
addHook("MobjDamage", JES.MobjDamage)
addHook("MobjDeath", JES.MobjDeath)

addHook("MobjThinker", JES.SliceJunkThinker, MT_JANA_SLICEJUNK)
addHook("MobjMoveBlocked", JES.SliceJunkMoveBlocked, MT_JANA_SLICEJUNK)

addHook("MobjThinker", JES.SaberHitboxThinker, MT_JANA_SABERHITBOX)
addHook("MobjCollide", JES.SaberHitboxCollide, MT_JANA_SABERHITBOX)
addHook("MobjMoveCollide", JES.SaberHitboxCollide, MT_JANA_SABERHITBOX)
addHook("MobjLineCollide", JES.SaberHitboxLineCollide, MT_JANA_SABERHITBOX)

addHook("MobjThinker", JES.SaberBeamThinker, MT_JANA_SMALLSABERBEAM)
addHook("MobjCollide", JES.SmallSaberBeamCollide, MT_JANA_SMALLSABERBEAM)
addHook("MobjMoveCollide", JES.SmallSaberBeamCollide, MT_JANA_SMALLSABERBEAM)
addHook("MobjLineCollide", JES.SaberBeamLineCollide, MT_JANA_SMALLSABERBEAM)

addHook("MobjThinker", JES.SaberBeamThinker, MT_JANA_LARGESABERBEAM)
addHook("MobjCollide", JES.LargeSaberBeamCollide, MT_JANA_LARGESABERBEAM)
addHook("MobjMoveCollide", JES.LargeSaberBeamCollide, MT_JANA_LARGESABERBEAM)
addHook("MobjLineCollide", JES.SaberBeamLineCollide, MT_JANA_LARGESABERBEAM)

addHook("MobjThinker", JES.SaberBeamHitboxThinker, MT_JANA_LARGESABERBEAM_HITBOX)
addHook("MobjRemoved", JES.SaberBeamHitboxRemoved, MT_JANA_LARGESABERBEAM_HITBOX)
addHook("MobjCollide", JES.SaberBeamHitboxCollide, MT_JANA_LARGESABERBEAM_HITBOX)
addHook("MobjMoveCollide", JES.SaberBeamHitboxCollide, MT_JANA_LARGESABERBEAM_HITBOX)
addHook("MobjLineCollide", JES.SaberBeamHitboxLineCollide, MT_JANA_LARGESABERBEAM_HITBOX)

addHook("MobjCollide", JES.SpikeCollide, MT_SPIKE)
addHook("MobjCollide", JES.SpikeCollide, MT_WALLSPIKE)

addHook("MobjCollide", JES.EggShieldCollide, MT_EGGSHIELD)
addHook("MobjMoveCollide", JES.EggShieldCollide, MT_EGGSHIELD)

hud.add(JES.DebugHUD)